------------------------------------------------------------
-- Core.lua
--
-- Abin
-- 2011-11-13
------------------------------------------------------------

local type = type
local tinsert = tinsert
local GetSpellInfo = GetSpellInfo
local select = select
local UnitBuff = UnitBuff
local GetNumShapeshiftForms = GetNumShapeshiftForms
local GetShapeshiftFormInfo = GetShapeshiftFormInfo
local UnitName = UnitName
local UnitClass = UnitClass
local format = format
local pairs = pairs
local ipairs = ipairs
local tostring = tostring
local GetActiveSpecGroup = GetActiveSpecGroup
local GetRealmName = GetRealmName
local GetNumGroupMembers = GetNumGroupMembers
local IsInRaid = IsInRaid
local wipe = wipe
local RAID_CLASS_COLORS = RAID_CLASS_COLORS

local addonName, addon = ...
_G["LiteBuff"] = addon
addon.version = GetAddOnMetadata(addonName, "Version") or "2.0"

local actionButtons = {} -- All created action buttons
local InitCallbacks = {} -- Registered functions to be called when ADDON_LOADED fires for this addon, after addon data are intialized

------------------------------------------------------------
-- Creates an action button
--
-- key: string, the unique identifier for this button, i.e., "PaladinBlessings", "HunterPets", etc. whatever you like, please make sure it's unique among all buttons
-- category: string, category of this button, i.e., "blessing", "poison"
-- title: string, optional, title of the button to be displayed in the option page
-- duration: number/nil, estimated aura full duration used for text color determining
-- ...: flags, see definitions located in the "Templates" folder
------------------------------------------------------------
function addon:CreateActionButton(key, category, title, duration, ...)
	local button = self.templates.CreateActionButton(key, category, title, duration, ...)
	if button then
		tinsert(actionButtons, button)
		return button
	end
end

function addon:GetNumButtons()
	return #actionButtons
end

function addon:GetButton(index)
	if type(index) == "string" then
		local button
		for _, button in ipairs(actionButtons) do
			if button.key == index then
				return button
			end
		end
	else
		return actionButtons[index]
	end
end

-- Register a function which will be called when ADDON_LOADED fires for this addon
function addon:RegisterInitCallback(func, arg1)
	if InitCallbacks and type(func) == "function" then
		tinsert(InitCallbacks, { func = func, arg1 = arg1 } )
		return 1
	end
end

-- Builds a spell list using given spell id and conflicts list
function addon:BuildSpellList(spellList, spellId, ...)
	local spell, _, icon = GetSpellInfo(spellId)
	if not spell then
		return
	end

	local data = { id = spellId, spell = spell, icon = icon }
	local conflicts

	-- build conflicts list
	local i
	for i = 1, select("#", ...) do
		local id = select(i, ...)
		local spell, _, icon = GetSpellInfo(id)
		if spell then
			if not conflicts then
				conflicts = {}
			end
			conflicts[spell] = icon
		end
	end
	data.conflicts = conflicts

	if type(spellList) == "table" then
		tinsert(spellList, data)
	end
	return data
end

-- Retrieves buff remain time
function addon:GetUnitBuffTimer(unit, buff, mine)
	if unit and buff then
		local name, _, _, count, _, _, expires, caster = UnitBuff(unit, buff)
		if name and (not mine or caster == "player") then
			return expires or 0, count or 1
		end
	end
end

-- Generates a gradient color for text displaying
function addon:GetGradientColor(number, threshold)
	local r, g = 1, 0
	if number and threshold and threshold > 0 then
		local percent = number / threshold
		if percent >= 0.5 then
			r, g = (1.0 - percent) * 2, 1
		else
			r, g = 1, percent * 2
		end
	end
	return r, g, 0
end

-- Checks whether the player is under the given form/stance/shape-shift/paladin-aura
function addon:IsFormActive(form)
	local i
	for i = 1, GetNumShapeshiftForms() do
		local _, name, active = GetShapeshiftFormInfo(i)
		if name == form then
			return active
		end
	end
end

-- Decorates an unit name with its class color
function addon:GetColoredUnitName(unit)
	if not unit then
		return
	end

	local name = UnitName(unit)
	if not name then
		return
	end

	local color = RAID_CLASS_COLORS[select(2, UnitClass(unit))]
	if color then
		name = format("|cff%02x%02x%02x%s|r", color.r * 255, color.g * 255, color.b * 255, name)
	end
	return name
end

-- Checks wether the player is grouped (raid/party/nil)
function addon:IsGrouped()
	local count = GetNumGroupMembers()
	local group
	if IsInRaid() then
		group = "raid"
	elseif count > 0 then
		group = "party"
	end

	return group, count
end

-- Data access
local DATA_TABLES = { db = 1, chardb = 1, specdb = 1, disabledb = 1 }
local function GetDataTable(dataType)
	return DATA_TABLES[dataType] and addon[dataType]
end

function addon:LoadData(dataType, key)
	local data = GetDataTable(dataType)
	return data and data[key]
end

function addon:SaveData(dataType, key, value)
	local data = GetDataTable(dataType)
	if data then
		data[key] = value
		return 1
	end
end

-- Predefined game events and corresponding methods
local EVENTS_DEF = {
	UNIT_AURA =			{ method = "OnPlayerAura", arg1 = "player" },
	UNIT_INVENTORY_CHANGED =	{ method = "OnInventoryUpdate", arg1 = "player" },
	SPELLS_CHANGED =		{ method = "OnSpellUpdate", arg1 = "nil" },
	BAG_UPDATE =			{ method = "OnBagUpdate" },
	BAG_UPDATE_COOLDOWN =		{ method = "OnBagUpdate" },
	PLAYER_TALENT_UPDATE =		{ method = "OnTalentUpdate" },
	UNIT_STATS =			{ method = "OnStatsUpdate", arg1 = "player" },
	RAID_ROSTER_UPDATE =		{ method = "OnRosterUpdate" },
	PARTY_MEMBERS_CHANGED =		{ method = "OnRosterUpdate" },
	UNIT_PET =			{ method = "OnPlayerPet", arg1 = "player" },
	PLAYER_TOTEM_UPDATE =		{ method = "OnPlayerPet" },
}

local inCombat
local methodPool = {}

local function FireAllEvents()
	local data
	for _, data in pairs(EVENTS_DEF) do
		methodPool[data.method] = 1
	end
end

local function NotifyButtons(method)
	local i
	for i = 1, #actionButtons do
		actionButtons[i]:InvokeMethod(method, inCombat)
	end
end

local function OnTalentSwitch()
	local db = addon.chardb.talents
	if type(db) ~= "table" then
		db = {}
		addon.chardb.talents = db
	end

	local talent = GetActiveSpecGroup()
	if type(db[talent]) ~= "table" then
		db[talent] = {}
	end
	addon.specdb = db[talent]
	NotifyButtons("OnTalentSwitch")
end

local updateElapsed = 0
local function Frame_OnUpdate(self, elapsed)
	updateElapsed = updateElapsed + elapsed
	if updateElapsed > 0.2 then
		updateElapsed = 0
		local method
		for method in pairs(methodPool) do
			NotifyButtons(method)
			methodPool[method] = nil
		end
	end
end

--------------------------------------------
-- Addon main frame
--------------------------------------------

local frame = CreateFrame("Frame", "LiteBuffFrame", UIParent, "SecureFrameTemplate")
addon.frame = frame
frame:SetSize(75, 68)
frame:SetPoint("CENTER")
frame:SetMovable(true)
frame:SetClampedToScreen(true)
frame:SetUserPlaced(true)
frame:RegisterEvent("ADDON_LOADED")

frame:SetScript("OnEvent", function(self, event, arg1)
	if event == "ADDON_LOADED" and arg1 == addonName then
		self:UnregisterEvent(event)

		-- Initialize SavedVariables table
		if type(LiteBuffDB) ~= "table" then
			LiteBuffDB = {}
		end
		addon.db = LiteBuffDB

		-- Cleanup data from old versions
		if not addon.db.v20 then
			wipe(addon.db)
			addon.db.v20 = 1
		end

		if type(LiteBuffCharDB) ~= "table" then
			LiteBuffCharDB = {}
		end
		addon.chardb = LiteBuffCharDB

		if type(addon.chardb.disabled) ~= "table" then
			addon.chardb.disabled = {}
		end

		addon.disabledb = addon.chardb.disabled

		-- Call "OnInitialize" for all buttons immediately
		NotifyButtons("OnInitialize")

		-- Invoke functions registered from addon:RegisterInitCallback
		local key, data
		for key, data in ipairs(InitCallbacks) do
			data.func(data.arg1)
		end
		InitCallbacks = nil

		self:RegisterEvent("PLAYER_ENTERING_WORLD")
		self:RegisterEvent("PLAYER_REGEN_DISABLED")
		self:RegisterEvent("PLAYER_REGEN_ENABLED")
		self:RegisterEvent("ACTIVE_TALENT_GROUP_CHANGED")

		-- Register predefined game events
		local key
		for key in pairs(EVENTS_DEF) do
			self:RegisterEvent(key)
		end

		OnTalentSwitch()
		self:SetScript("OnUpdate", Frame_OnUpdate)

	elseif event == "PLAYER_ENTERING_WORLD" then
		FireAllEvents()
		NotifyButtons("OnEnterWorld")

	elseif event == "PLAYER_REGEN_DISABLED" then
		inCombat = 1
		NotifyButtons("OnEnterCombat")

	elseif event == "PLAYER_REGEN_ENABLED" then
		inCombat = nil
		FireAllEvents()
		NotifyButtons("OnLeaveCombat")
		Frame_OnUpdate(self, 1000) -- immediately invoke all methods upon leaving combat, no delays applied

	elseif event == "ACTIVE_TALENT_GROUP_CHANGED" then
		OnTalentSwitch()

	else
		-- Invoke methods defined in table EVENTS_DEF
		local data = EVENTS_DEF[event]
		if data then
			if not data.arg1 or data.arg1 == tostring(arg1) then
				methodPool[data.method] = 1
			end
		end
	end
end)